/*********************
Yusheng Yang
Real-Time Innovations
August 4th, 2008
**********************/

/****************************************************************************************
Status_PrettyPrint_API is a static class for printing reader/writer statistics to a .csv file.
*****************************************************************************************/

#include <stdlib.h>
#include <fstream>
#include <iostream>
#include <string>
#include <stdio.h>
#include "ndds/ndds_cpp.h"
#include "Status_PrettyPrint_API.h"

/*********************************************
Appends a reader's status/statistics to a
.csv file named after the reader's GUID.
If the file doesn't exist, it creates a new
file.

Input: reader whose status is to be printed
Output: return code indicating success/failure
**********************************************/

DDS_ReturnCode_t Status_PrettyPrint_API::print_reader_status_to_csv (DDSDataReader *reader) {
	if (reader == NULL) {
		printf("print_reader_status_to_csv: non-null reader required\n");
		return DDS_RETCODE_ERROR;
	}

	/* if filename is not specified, use "reader_<guid>_status.csv" */
	string csvFile = "reader_";
	string guidString;

	DDS_ReturnCode_t retcode = Status_PrettyPrint_API::get_reader_guid_string(reader, guidString);
	if (retcode != DDS_RETCODE_OK) return retcode;

	csvFile += guidString + "_status.csv";

	return Status_PrettyPrint_API::print_reader_status_to_csv(reader, csvFile.c_str());
}

/*********************************************
Appends a reader's status/statistics to a
specified .csv file.
If the file doesn't exist, it creates a new
file.

Inputs:
 reader - reader whose status is to be printed
 csvFile - name of the .csv file to print to
Output: return code indicating success/failure
**********************************************/

DDS_ReturnCode_t Status_PrettyPrint_API::print_reader_status_to_csv (DDSDataReader *reader, const char csvFile[]) {
	if (reader == NULL) {
		printf("print_reader_status_to_csv: non-null reader required\n");
		return DDS_RETCODE_ERROR;
	}
	if (csvFile == NULL) return Status_PrettyPrint_API::print_reader_status_to_csv(reader);

	FILE *f;	
	DDS_ReturnCode_t retcode;

	/* if the file doesn't exist, create a new file and print its title/header info */
	if ((f = fopen(csvFile, "r")) == NULL) {
		retcode = Status_PrettyPrint_API::init_reader_status_csv(reader, csvFile);
		if (retcode != DDS_RETCODE_OK) return retcode;
	} else {
		if (fclose(f) != 0)
			return DDS_RETCODE_ERROR;
	}

	/* open the file */
	ofstream outfile(csvFile, ios::app);
	if (!outfile.is_open())	{
		printf("print_reader_status_to_csv: open csvFile error\n");
		return DDS_RETCODE_ERROR;
	}	

	/* print timestamp */
	retcode = Status_PrettyPrint_API::print_timestamp(reader->get_subscriber()->get_participant(), outfile);
	if (retcode != DDS_RETCODE_OK) return retcode;

	/* print cache status */	
	retcode = Status_PrettyPrint_API::print_reader_cache_status(reader, outfile);
	if (retcode != DDS_RETCODE_OK) return retcode;

	/* print protocol status */
	retcode = Status_PrettyPrint_API::print_reader_protocol_status(reader, outfile);
	if (retcode != DDS_RETCODE_OK) return retcode;

	outfile << endl;

	/* close the file */
	outfile.close();

	return DDS_RETCODE_OK;
}

/*********************************************
Appends a writer's status/statistics to a
.csv file named after the writer's GUID.
If the file doesn't exist, it creates a new
file.

Input: writer whose status is to be printed
Output: return code indicating success/failure
**********************************************/

DDS_ReturnCode_t Status_PrettyPrint_API::print_writer_status_to_csv (DDSDataWriter *writer) {
	if (writer == NULL) {
		printf("print_writer_status_to_csv: non-null writer required\n");
		return DDS_RETCODE_ERROR;
	}

	/* if filename is not specified, use "writer_<guid>_status.csv" */
	string csvFile = "writer_";
	string guidString;

	DDS_ReturnCode_t retcode = Status_PrettyPrint_API::get_writer_guid_string(writer, guidString);
	if (retcode != DDS_RETCODE_OK) return retcode;

	csvFile += guidString + "_status.csv";

	return Status_PrettyPrint_API::print_writer_status_to_csv(writer, csvFile.c_str());
}

/*********************************************
Appends a writer's status/statistics to a
specified .csv file.
If the file doesn't exist, it creates a new
file.

Inputs:
 writer - writer whose status is to be printed
 csvFile - name of the .csv file to print to
Output: return code indicating success/failure
**********************************************/

DDS_ReturnCode_t Status_PrettyPrint_API::print_writer_status_to_csv (DDSDataWriter *writer, const char csvFile[]) {
	if (writer == NULL) {
		printf("print_writer_status_to_csv: non-null writer required\n");
		return DDS_RETCODE_ERROR;
	}
	if (csvFile == NULL) return Status_PrettyPrint_API::print_writer_status_to_csv(writer);

	FILE *f;	
	DDS_ReturnCode_t retcode;

	/* if the file doesn't exist, create a new file and print its title/header info */
	if ((f = fopen(csvFile, "r")) == NULL) {
		retcode = Status_PrettyPrint_API::init_writer_status_csv(writer, csvFile);
		if (retcode != DDS_RETCODE_OK) return retcode;
	} else {
		if (fclose(f) != 0)
			return DDS_RETCODE_ERROR;
	}

	/* open the file */	
	ofstream outfile(csvFile, ios::app);
	if (!outfile.is_open())	{
		printf("print_writer_status_to_csv: open csvFile error\n");
		return DDS_RETCODE_ERROR;
	}

	/* print timestamp */
	retcode = Status_PrettyPrint_API::print_timestamp(writer->get_publisher()->get_participant(), outfile);
	if (retcode != DDS_RETCODE_OK) return retcode;

	/* print cache status */	
	retcode = Status_PrettyPrint_API::print_writer_cache_status(writer, outfile);
	if (retcode != DDS_RETCODE_OK) return retcode;

	/* print protocol status */
	retcode = Status_PrettyPrint_API::print_writer_protocol_status(writer, outfile);
	if (retcode != DDS_RETCODE_OK) return retcode;

	outfile << endl;

	/* close the file */
	outfile.close();

	return DDS_RETCODE_OK;
}

/*********************************************
Initializes a reader's status/statistics file
by printing the title and column headers. The
.csv file is named after the reader's GUID.
If the file doesn't exist, it creates a new
file.

Input: reader whose status is to be printed
Output: return code indicating success/failure
**********************************************/

DDS_ReturnCode_t Status_PrettyPrint_API::init_reader_status_csv(DDSDataReader *reader) {
	if (reader == NULL) {
		printf("init_reader_status_csv: non-null reader required\n");
		return DDS_RETCODE_ERROR;
	}

	/* if filename is not specified, use "reader_<guid>_status.csv" */
	string csvFile = "reader_";
	string guidString;

	DDS_ReturnCode_t retcode = Status_PrettyPrint_API::get_reader_guid_string(reader, guidString);
	if (retcode != DDS_RETCODE_OK) return retcode;

	csvFile += guidString + "_status.csv";

	return Status_PrettyPrint_API::init_reader_status_csv(reader, csvFile.c_str());
}

/*********************************************
Initializes a reader's status/statistics file
by printing the title and column headers.
If the file doesn't exist, it creates a new
file.

Inputs:
 reader - reader whose status is to be printed
 csvFile - name of the .csv file to print to
Output: return code indicating success/failure
**********************************************/

DDS_ReturnCode_t Status_PrettyPrint_API::init_reader_status_csv(DDSDataReader *reader, const char csvFile[]) {
	if (reader == NULL) {
		printf("init_reader_status_csv: non-null reader required\n");
		return DDS_RETCODE_ERROR;
	}
	if (csvFile == NULL) return Status_PrettyPrint_API::init_reader_status_csv(reader);

	/* open the file */
	ofstream outfile(csvFile);
	if (!outfile.is_open())	{
		printf("init_reader_status_csv: open csvFile error\n");
		return DDS_RETCODE_ERROR;
	}

	/* print the title */	
	DDS_ReturnCode_t retcode = Status_PrettyPrint_API::print_reader_status_title(reader, outfile);
	if (retcode != DDS_RETCODE_OK) return retcode;

	/* print the column headers */
	print_reader_column_headers(reader, outfile);

	/* close the file */
	outfile.close();

	return DDS_RETCODE_OK;
}

/*********************************************
Initializes a writer's status/statistics file
by printing the title and column headers. The
.csv file is named after the reader's GUID.
If the file doesn't exist, it creates a new
file.

Input: writer whose status is to be printed
Output: return code indicating success/failure
**********************************************/

DDS_ReturnCode_t Status_PrettyPrint_API::init_writer_status_csv(DDSDataWriter *writer) {
	if (writer == NULL) {
		printf("init_writer_status_csv: non-null writer required\n");
		return DDS_RETCODE_ERROR;
	}

	/* if filename is not specified, use "reader_<guid>_status.csv" */
	string csvFile = "writer_";
	string guidString;

	DDS_ReturnCode_t retcode = Status_PrettyPrint_API::get_writer_guid_string(writer, guidString);
	if (retcode != DDS_RETCODE_OK) return retcode;

	csvFile += guidString + "_status.csv";

	return Status_PrettyPrint_API::init_writer_status_csv(writer, csvFile.c_str());
}

/*********************************************
Initializes a writer's status/statistics file
by printing the title and column headers.
If the file doesn't exist, it creates a new
file.

Inputs:
 writer - writer whose status is to be printed
 csvFile - name of the .csv file to print to
Output: return code indicating success/failure
**********************************************/

DDS_ReturnCode_t Status_PrettyPrint_API::init_writer_status_csv(DDSDataWriter *writer, const char csvFile[]) {
	if (writer == NULL) {
		printf("init_writer_status_csv: non-null writer required\n");
		return DDS_RETCODE_ERROR;
	}
	if (csvFile == NULL) return Status_PrettyPrint_API::init_writer_status_csv(writer);

	/* open the file */
	ofstream outfile(csvFile);
	if (!outfile.is_open())	{
		printf("init_writer_status_csv: open csvFile error\n");
		return DDS_RETCODE_ERROR;
	}

	/* print the title */	
	DDS_ReturnCode_t retcode = Status_PrettyPrint_API::print_writer_status_title(writer, outfile);
	if (retcode != DDS_RETCODE_OK) return retcode;

	/* print the column headers */
	print_writer_column_headers(writer, outfile);	

	/* close the file */
	outfile.close();
	
	return DDS_RETCODE_OK;
}

/*** private methods ***/

DDS_ReturnCode_t Status_PrettyPrint_API::print_timestamp(DDSDomainParticipant *participant, ofstream &outfile) {
	DDS_Time_t time;

	DDS_ReturnCode_t retcode = participant->get_current_time(time);
	if (retcode != DDS_RETCODE_OK) return retcode;

	outfile << time.sec << "," << time.nanosec << ",";

	return DDS_RETCODE_OK;
}

DDS_ReturnCode_t Status_PrettyPrint_API::print_reader_status_title(DDSDataReader *reader, ofstream &outfile) {
	string guidString;

	DDS_ReturnCode_t retcode = Status_PrettyPrint_API::get_reader_guid_string(reader, guidString);
	if (retcode != DDS_RETCODE_OK) return retcode;

	outfile << "reader GUID:," << guidString << endl << endl;
		
	return DDS_RETCODE_OK;
}

void Status_PrettyPrint_API::print_reader_column_headers(DDSDataReader *reader, ofstream &outfile) {
	outfile << ",,CACHE STATUS,,PROTOCOL STATUS" << endl;
	outfile << "sec," << "nanosec,";
	outfile << "sample_count_peak," << "sample_count,";
	outfile << "received_sample_count," << "received_sample_count_change," << "received_sample_bytes," << "received_sample_bytes_change,";
	outfile << "duplicate_sample_count," << "duplicate_sample_count_change," << "duplicate_sample_bytes," << "duplicate_sample_bytes_change,";
	outfile << "filtered_sample_count," << "filtered_sample_count_change," << "filtered_sample_bytes," << "filtered_sample_bytes_change,";
	outfile << "received_heartbeat_count," << "received_heartbeat_count_change," << "received_heartbeat_bytes," << "received_heartbeat_bytes_change,";
	outfile << "sent_ack_count," << "sent_ack_count_change," << "sent_ack_bytes," << "sent_ack_bytes_change,";
	outfile << "sent_nack_count," << "sent_nack_count_change," << "sent_nack_bytes," << "sent_nack_bytes_change,";
	outfile << "received_gap_count," << "received_gap_count_change," << "received_gap_bytes," << "received_gap_bytes_change,";
	outfile << "rejected_sample_count," << "rejected_sample_count_change" << endl;
}

DDS_ReturnCode_t Status_PrettyPrint_API::print_reader_cache_status(DDSDataReader *reader, ofstream &outfile) {
	DDS_DataReaderCacheStatus status;

	DDS_ReturnCode_t retcode = reader->get_datareader_cache_status(status);
	if (retcode != DDS_RETCODE_OK) {
		printf("print_reader_cache_status: get_datareader_cache_status error\n");
		return DDS_RETCODE_ERROR;
	}

	/* print the values */
	outfile << status.sample_count_peak << "," << status.sample_count << ",";

	return DDS_RETCODE_OK;
}

DDS_ReturnCode_t Status_PrettyPrint_API::print_reader_protocol_status(DDSDataReader *reader, ofstream &outfile) {
	DDS_DataReaderProtocolStatus status;

	DDS_ReturnCode_t retcode = reader->get_datareader_protocol_status(status);
	if (retcode != DDS_RETCODE_OK) {
		printf("print_reader_protocol_status: get_datareader_protocol_status error\n");
		return DDS_RETCODE_ERROR;
	}

	/* print the values */
	outfile << status.received_sample_count << "," << status.received_sample_count_change << "," << status.received_sample_bytes << "," << status.received_sample_bytes_change << ",";		
	outfile << status.duplicate_sample_count << "," << status.duplicate_sample_count_change << "," << status.duplicate_sample_bytes << "," << status.duplicate_sample_bytes_change << ",";
	outfile << status.filtered_sample_count << "," << status.filtered_sample_count_change << "," << status.filtered_sample_bytes << "," << status.filtered_sample_bytes_change << ",";
	outfile << status.received_heartbeat_count << "," << status.received_heartbeat_count_change << "," << status.received_heartbeat_bytes << "," << status.received_heartbeat_bytes_change << ",";
	outfile << status.sent_ack_count << "," << status.sent_ack_count_change << "," << status.sent_ack_bytes << "," << status.sent_ack_bytes_change << ",";
	outfile << status.sent_nack_count << "," << status.sent_nack_count_change << "," << status.sent_nack_bytes << "," << status.sent_nack_bytes_change << ",";
	outfile << status.received_gap_count << "," << status.received_gap_count_change << "," << status.received_gap_bytes << "," << status.received_gap_bytes_change << ",";
	outfile << status.rejected_sample_count << "," << status.rejected_sample_count_change << ",";
	
	return DDS_RETCODE_OK;
}

DDS_ReturnCode_t Status_PrettyPrint_API::print_writer_status_title(DDSDataWriter *writer, ofstream &outfile) {		
	string guidString;

	DDS_ReturnCode_t retcode = Status_PrettyPrint_API::get_writer_guid_string(writer, guidString);
	if (retcode != DDS_RETCODE_OK) return retcode;

	outfile << "writer GUID:," << guidString << endl << endl;
		
	return DDS_RETCODE_OK;
}

void Status_PrettyPrint_API::print_writer_column_headers(DDSDataWriter *writer, ofstream &outfile) {
	outfile << ",,CACHE STATUS,,PROTOCOL STATUS" << endl;
	outfile << "sec," << "nanosec,";	
	outfile << "sample_count_peak," << "sample_count,";
	outfile << "pushed_sample_count," << "pushed_sample_count_change," << "pushed_sample_bytes," << "pushed_sample_bytes_change,";
	outfile << "filtered_sample_count," << "filtered_sample_count_change," << "filtered_sample_bytes," << "filtered_sample_bytes_change,";
	outfile << "sent_heartbeat_count," << "sent_heartbeat_count_change," << "sent_heartbeat_bytes," << "sent_heartbeat_bytes_change,";
	outfile << "pulled_sample_count," << "pulled_sample_count_change," << "pulled_sample_bytes," << "pulled_sample_bytes_change,";
	outfile << "received_ack_count," << "received_ack_count_change," << "received_ack_bytes," << "received_ack_bytes_change,";
	outfile << "received_nack_count," << "received_nack_count_change," << "received_nack_bytes," << "received_nack_bytes_change,";
	outfile << "sent_gap_count," << "sent_gap_count_change," << "sent_gap_bytes," << "sent_gap_bytes_change,";
	outfile << "rejected_sample_count," << "rejected_sample_count_change" << endl;
}

DDS_ReturnCode_t Status_PrettyPrint_API::print_writer_cache_status(DDSDataWriter *writer, ofstream &outfile) {
	DDS_DataWriterCacheStatus status;

	DDS_ReturnCode_t retcode = writer->get_datawriter_cache_status(status);
	if (retcode != DDS_RETCODE_OK) {
		printf("print_writer_cache_status: get_datawriter_cache_status error\n");
		return DDS_RETCODE_ERROR;
	}	

	/* print the values */
	outfile << status.sample_count_peak << "," << status.sample_count << ",";

	return DDS_RETCODE_OK;
}

DDS_ReturnCode_t Status_PrettyPrint_API::print_writer_protocol_status(DDSDataWriter *writer, ofstream &outfile) {
	DDS_DataWriterProtocolStatus status;

	DDS_ReturnCode_t retcode = writer->get_datawriter_protocol_status(status);
	if (retcode != DDS_RETCODE_OK) {
		printf("print_writer_protocol_status: get_datawriter_protocol_status error\n");
		return DDS_RETCODE_ERROR;
	}	

	/* print the values */	
	outfile << status.pushed_sample_count << "," << status.pushed_sample_count_change << "," << status.pushed_sample_bytes << "," << status.pushed_sample_bytes_change << ",";	
	outfile << status.filtered_sample_count << "," << status.filtered_sample_count_change << "," << status.filtered_sample_bytes << "," << status.filtered_sample_bytes_change << ",";	
	outfile << status.sent_heartbeat_count << "," << status.sent_heartbeat_count_change << "," << status.sent_heartbeat_bytes << "," << status.sent_heartbeat_bytes_change << ",";	
	outfile << status.pulled_sample_count << "," << status.pulled_sample_count_change << "," << status.pulled_sample_bytes << "," << status.pulled_sample_bytes_change << ",";	
	outfile << status.received_ack_count << "," << status.received_ack_count_change << "," << status.received_ack_bytes << "," << status.received_ack_bytes_change << ",";	
	outfile << status.received_nack_count << "," << status.received_nack_count_change << "," << status.received_nack_bytes << "," << status.received_nack_bytes_change << ",";	
	outfile << status.sent_gap_count << "," << status.sent_gap_count_change << "," << status.sent_gap_bytes << "," << status.sent_gap_bytes_change << ",";	
	outfile << status.rejected_sample_count << "," << status.rejected_sample_count_change << ",";
	
	return DDS_RETCODE_OK;
}

DDS_ReturnCode_t Status_PrettyPrint_API::get_reader_guid_string(DDSDataReader *reader, string &reader_guid_string) {
	DDS_DataReaderQos qos;

	DDS_ReturnCode_t retcode = reader->get_qos(qos);
	if (retcode != DDS_RETCODE_OK) {
		printf("get_reader_guid_string: get_qos error\n");
		return DDS_RETCODE_ERROR;
	}

	reader_guid_string = "";
	
	for (int i = 0; i < GUID_LENGTH; i++)		
		reader_guid_string += DDSOctet2hexStr(qos.protocol.virtual_guid.value[i]);	

	return DDS_RETCODE_OK;	
}

DDS_ReturnCode_t Status_PrettyPrint_API::get_writer_guid_string(DDSDataWriter *writer, string &writer_guid_string) {
	DDS_DataWriterQos qos;	 

	DDS_ReturnCode_t retcode = writer->get_qos(qos);
	if (retcode != DDS_RETCODE_OK) {
		printf("get_writer_guid_string: get_qos error\n");
		return DDS_RETCODE_ERROR;
	}

	writer_guid_string = "";
		
	for (int i = 0; i < GUID_LENGTH; i++)
		writer_guid_string += DDSOctet2hexStr(qos.protocol.virtual_guid.value[i]);	
		
	return DDS_RETCODE_OK;
}

string Status_PrettyPrint_API::DDSOctet2hexStr(DDS_Octet oct) {
	char result[2];
	char lower = oct % 16;
	char upper = oct >> 4;

	if (lower < 10) result[1] = lower + '0';
	else result[1] = lower % 10 + 'a';
	if (upper < 10) result[0] = upper + '0';
	else result[0] = upper % 10 + 'a';
	return string(result, 2);
}
